# -*- coding: utf-8 -*-
import django

import pytest
from demo.base import SENTINEL, AdminTestCase
from demo.models import SimpleConcurrentModel
from demo.util import unique_id


@pytest.mark.xfail("django.VERSION[:2] == (1, 10)", strict=True)
class TestAdminActions(AdminTestCase):
    def _create_conflict(self, pk):
        u = SimpleConcurrentModel.objects.get(pk=pk)
        u.username = SENTINEL
        u.save()

    def test_dummy_action(self):
        id = next(unique_id)
        SimpleConcurrentModel.objects.get_or_create(pk=id)
        res = self.app.get('/admin/', user='sax')

        res = res.click('^SimpleConcurrentModels')
        assert 'SimpleConcurrentModel #%s' % id in res  # sanity check

        self._create_conflict(id)

        form = res.forms['changelist-form']
        form['action'].value = 'dummy_action'
        sel = form.get('_selected_action', index=0)
        sel.checked = True
        res = form.submit().follow()

        self.assertIn('SimpleConcurrentModel #%s' % id, res)
        self.assertIn('**concurrent_update**', res)
        self.assertNotIn('**action_update**', res)

    def test_dummy_action_select_across(self):
        id = next(unique_id)
        SimpleConcurrentModel.objects.get_or_create(pk=id)
        res = self.app.get('/admin/', user='sax')

        res = res.click('^SimpleConcurrentModels')
        assert 'SimpleConcurrentModel #%s' % id in res  # sanity check

        self._create_conflict(id)

        form = res.forms['changelist-form']
        form['action'].value = 'dummy_action'
        form['select_across'] = 'True'
        sel = form.get('_selected_action', index=0)  # needed
        sel.checked = True  # needed
        res = form.submit()
        res = res.follow()

        self.assertIn('Selecting all records, you will avoid the concurrency check', res)

    # @pytest.mark.skipif(django.VERSION[:2] >= (1, 7), reason="Skip django>=1.9")
    def test_delete_allowed_if_no_updates(self):
        id = next(unique_id)
        SimpleConcurrentModel.objects.get_or_create(pk=id)
        res = self.app.get('/admin/', user='sax')
        res = res.click('^SimpleConcurrentModels')
        assert 'SimpleConcurrentModel #%s' % id in res  # sanity check

        form = res.forms['changelist-form']
        form['action'].value = 'delete_selected'
        sel = form.get('_selected_action', index=0)
        sel.checked = True

        res = form.submit()

        assert 'Are you sure?' in res
        assert 'SimpleConcurrentModel #%s' % id in res
        res = res.form.submit()
        assert 'SimpleConcurrentModel #%s' % id not in res

    # @pytest.mark.skipif(django.VERSION[:2] >= (1, 10), reason="Skip django>=1.10")
    def test_delete_not_allowed_if_updates(self):
        id = next(unique_id)

        SimpleConcurrentModel.objects.get_or_create(pk=id)
        res = self.app.get('/admin/', user='sax')

        res = res.click('^SimpleConcurrentModels')
        assert 'SimpleConcurrentModel #%s' % id in res  # sanity check

        self._create_conflict(id)

        form = res.forms['changelist-form']
        form['action'].value = 'delete_selected'
        sel = form.get('_selected_action', index=0)
        sel.checked = True
        res = form.submit().follow()
        self.assertIn('One or more record were updated', res)

    @pytest.mark.django_db
    def test_deleteaction(self):
        id = next(unique_id)

        SimpleConcurrentModel.objects.get_or_create(pk=id)
        response = self.app.get(django.core.urlresolvers.reverse('admin:demo_simpleconcurrentmodel_changelist'),
                                user='sax')
        form = response.forms['changelist-form']
        form.get('_selected_action', index=0).checked = True
        form['action'] = 'delete_selected'
        response = form.submit()
        expected = 'All of the following objects and their related items will be deleted'
        assert expected in response
        response = response.form.submit().follow()
        assert response.status_code == 200



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models

import concurrency.fields


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('demo', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ConditionalVersionModelWithoutMeta',
            fields=[
                ('id', models.AutoField(auto_created=True, verbose_name='ID', serialize=False, primary_key=True)),
                ('version', concurrency.fields.ConditionalVersionField(default=1, help_text='record revision number')),
                ('field1', models.CharField(unique=True, blank=True, max_length=30, null=True)),
                ('field2', models.CharField(unique=True, blank=True, max_length=30, null=True)),
                ('field3', models.CharField(unique=True, blank=True, max_length=30, null=True)),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL, null=True)),
            ],
        ),
    ]



from django.conf.urls import include, url
from django.contrib import admin
from django.views.generic.edit import UpdateView

from demo.models import SimpleConcurrentModel

# try:
# from django.apps import AppConfig  # noqa
#     import django
#     django.setup()
# except ImportError:
#     pass

admin.autodiscover()

urlpatterns = (url('cm/(?P<pk>\d+)/',
                   UpdateView.as_view(model=SimpleConcurrentModel),
                   name='concurrent-edit'),
               url(r'^admin/',
                   include(admin.site.urls))
               )



import os
import platform
import sys

import django

import pytest

py_impl = getattr(platform, 'python_implementation', lambda: None)
PYPY = py_impl() == 'PyPy'
PURE_PYTHON = os.environ.get('PURE_PYTHON', False)

windows = pytest.mark.skipif(sys.platform != 'win32', reason="requires windows")

win32only = pytest.mark.skipif("sys.platform != 'win32'")

skippypy = pytest.mark.skipif(PYPY, reason='skip on pypy')

# skipIfDjangoVersion = lambda v: pytest.mark.skipif("django.VERSION[:2]>={}".format(v),
#                                                    reason="Skip if django>={}".format(v))


def pytest_configure():
    from django.contrib.auth.models import Group
    from django.conf import settings

    settings.SILENCED_SYSTEM_CHECKS = ['concurrency.W001']
    if django.VERSION[:2] == (1.6):
        from concurrency.api import apply_concurrency_check
        from concurrency.fields import IntegerVersionField

        apply_concurrency_check(Group, 'version', IntegerVersionField)


@pytest.fixture(scope='session')
def client(request):
    import django_webtest

    wtm = django_webtest.WebTestMixin()
    wtm.csrf_checks = False
    wtm._patch_settings()
    request.addfinalizer(wtm._unpatch_settings)
    app = django_webtest.DjangoTestApp()
    return app



# -*- coding: utf-8 -*-
from django.conf import settings
from django.contrib.admin.sites import site
from django.core.urlresolvers import reverse
from django.http import HttpRequest
from django.test.utils import override_settings

import mock
import pytest
from demo.base import AdminTestCase
from demo.models import SimpleConcurrentModel
from demo.util import DELETE_ATTRIBUTE, attributes, unique_id

from concurrency.admin import ConcurrentModelAdmin
from concurrency.config import CONCURRENCY_LIST_EDITABLE_POLICY_ABORT_ALL
from concurrency.exceptions import RecordModifiedError
from concurrency.middleware import ConcurrencyMiddleware


def _get_request(path):
    request = HttpRequest()
    request.META = {
        'SERVER_NAME': 'testserver',
        'SERVER_PORT': 80,
    }
    request.path = request.path_info = "/middleware/%s" % path
    return request


def test_middleware():
    handler = mock.Mock(status_code=409)
    type(handler.return_value).status_code = mock.PropertyMock(return_value=409)

    with override_settings(CONCURRENCY_HANDLER409=handler):
        request = _get_request('needsquoting#')
        r = ConcurrencyMiddleware().process_exception(request, RecordModifiedError(target=SimpleConcurrentModel()))
    assert r.status_code == 409


class ConcurrencyMiddlewareTest(AdminTestCase):
    def _get_request(self, path):
        request = HttpRequest()
        request.META = {
            'SERVER_NAME': 'testserver',
            'SERVER_PORT': 80,
        }
        request.path = request.path_info = "/middleware/%s" % path
        return request

    @mock.patch('django.core.signals.got_request_exception.send', mock.Mock())
    def test_process_exception(self):
        """
        Tests that RecordModifiedError is handled correctly.
        """
        id = next(unique_id)
        m, __ = SimpleConcurrentModel.objects.get_or_create(pk=id)
        copy = SimpleConcurrentModel.objects.get(pk=m.pk)
        copy.save()
        request = self._get_request('/')
        r = ConcurrencyMiddleware().process_exception(request, RecordModifiedError(target=m))
        self.assertEqual(r.status_code, 409)

    @pytest.mark.xfail("django.VERSION[:2] == (1, 10)", strict=True)
    def test_in_admin(self):
        id = next(unique_id)

        middlewares = list(settings.MIDDLEWARE_CLASSES) + ['concurrency.middleware.ConcurrencyMiddleware']
        model_admin = site._registry[SimpleConcurrentModel]

        with attributes((model_admin.__class__, 'list_editable_policy', CONCURRENCY_LIST_EDITABLE_POLICY_ABORT_ALL),
                        (ConcurrentModelAdmin, 'form', DELETE_ATTRIBUTE)):

            with self.settings(MIDDLEWARE_CLASSES=middlewares):

                saved, __ = SimpleConcurrentModel.objects.get_or_create(pk=id)

                url = reverse('admin:demo_simpleconcurrentmodel_change', args=[saved.pk])
                res = self.app.get(url, user='sax')
                form = res.form

                saved.save()  # create conflict here

                res = form.submit(expect_errors=True)

                self.assertEqual(res.status_code, 409)

                target = res.context['target']
                self.assertIn('target', res.context)
                self.assertIn('saved', res.context)

                self.assertEqual(res.context['target'].version, target.version)
                self.assertEqual(res.context['saved'].version, saved.version)
                self.assertEqual(res.context['request_path'], url)

#
# class TestFullStack(DjangoAdminTestCase):
#     MIDDLEWARE_CLASSES = ('django.middleware.common.CommonMiddleware',
#                           'django.contrib.sessions.middleware.SessionMiddleware',
#                           'django.contrib.auth.middleware.AuthenticationMiddleware',
#                           'django.contrib.messages.middleware.MessageMiddleware',
#                           'concurrency.middleware.ConcurrencyMiddleware',)
#
#     @mock.patch('django.core.signals.got_request_exception.send', mock.Mock())
#     def test_stack(self):
#         admin_register(TestModel0, ModelAdmin)
#
#         with self.settings(MIDDLEWARE_CLASSES=self.MIDDLEWARE_CLASSES):
#             m, __ = TestModel0.objects.get_or_create(username="New", last_name="1")
#             copy = TestModel0.objects.get(pk=m.pk)
#             assert copy.version == m.version
#             print 111111111111, m.version
#             url = reverse('admin:concurrency_testmodel0_change', args=[m.pk])
#             data = {'username': 'new_username',
#                     'last_name': None,
#                     'version': VersionFieldSigner().sign(m.version),
#                     'char_field': None,
#                     '_continue': 1,
#                     'date_field': '2010-09-01'}
#             copy.save()
#             assert copy.version > m.version
#
#             r = self.client.post(url, data, follow=True)
#             self.assertEqual(r.status_code, 409)
#             self.assertIn('target', r.context)
#             self.assertIn('saved', r.context)
#             self.assertEqual(r.context['saved'].version, copy.version)
#             self.assertEqual(r.context['target'].version, m.version)
#             self.assertEqual(r.context['request_path'], url)



from django.core.urlresolvers import reverse
from django.utils.translation import ugettext as _

import pytest
from demo.base import SENTINEL, AdminTestCase
from demo.models import SimpleConcurrentModel
from demo.util import nextname

from concurrency.forms import VersionFieldSigner


@pytest.mark.django_db
@pytest.mark.admin
def test_creation(admin_user, client):
    url = reverse('admin:demo_simpleconcurrentmodel_add')
    res = client.get(url, user=admin_user.username)

    form = res.form
    form['username'] = 'CHAR'
    res = form.submit().follow()
    assert SimpleConcurrentModel.objects.filter(username='CHAR').exists()
    assert SimpleConcurrentModel.objects.get(username='CHAR').version > 0


@pytest.mark.django_db
@pytest.mark.functional
def test_standard_update(admin_user, client):
    concurrentmodel = SimpleConcurrentModel.objects.create()
    url = reverse('admin:demo_simpleconcurrentmodel_change',
                  args=[concurrentmodel.pk])
    res = client.get(url, user=admin_user.username)

    target = res.context['original']

    old_version = target.version
    form = res.form
    form['username'] = 'UPDATED'
    res = form.submit().follow()
    target = SimpleConcurrentModel.objects.get(pk=target.pk)
    new_version = target.version

    assert new_version > old_version


@pytest.mark.django_db
@pytest.mark.functional
def test_conflict(admin_user, client):
    concurrentmodel = SimpleConcurrentModel.objects.create()
    url = reverse('admin:demo_simpleconcurrentmodel_change',
                  args=[concurrentmodel.pk])
    res = client.get(url, user=admin_user.username)
    form = res.form
    concurrentmodel.save()  # create conflict here

    res = form.submit()

    assert 'original' in res.context
    assert res.context['adminform'].form.errors
    assert _('Record Modified') in str(res.context['adminform'].form.errors)


class TestConcurrentModelAdmin(AdminTestCase):
    def test_standard_update(self):
        target, __ = SimpleConcurrentModel.objects.get_or_create(username='aaa')
        url = reverse('admin:demo_simpleconcurrentmodel_change', args=[target.pk])
        res = self.app.get(url, user='sax')
        target = res.context['original']
        old_version = target.version
        form = res.form
        form['username'] = 'UPDATED'
        res = form.submit().follow()
        target = SimpleConcurrentModel.objects.get(pk=target.pk)
        new_version = target.version
        self.assertGreater(new_version, old_version)

    def test_creation(self):
        url = reverse('admin:demo_simpleconcurrentmodel_add')
        res = self.app.get(url, user='sax')
        form = res.form
        form['username'] = 'CHAR'
        res = form.submit().follow()
        self.assertTrue(SimpleConcurrentModel.objects.filter(username='CHAR').exists())
        self.assertGreater(SimpleConcurrentModel.objects.get(username='CHAR').version, 0)

    def test_conflict(self):
        target, __ = SimpleConcurrentModel.objects.get_or_create(username='aaa')
        url = reverse('admin:demo_simpleconcurrentmodel_change', args=[target.pk])
        res = self.app.get(url, user='sax')

        form = res.form
        target.save()  # create conflict here

        res = form.submit()

        self.assertIn('original', res.context)
        self.assertTrue(res.context['adminform'].form.errors,
                        res.context['adminform'].form.errors)
        self.assertIn(_('Record Modified'),
                      str(res.context['adminform'].form.errors),
                      res.context['adminform'].form.errors)


class TestAdminEdit(AdminTestCase):
    def _create_conflict(self, pk):
        u = SimpleConcurrentModel.objects.get(pk=pk)
        u.username = SENTINEL
        u.save()

    def test_creation(self):
        url = reverse('admin:demo_simpleconcurrentmodel_add')
        res = self.app.get(url, user='sax')
        form = res.form
        form['username'] = 'CHAR'
        res = form.submit().follow()
        self.assertTrue(SimpleConcurrentModel.objects.filter(username='CHAR').exists())
        self.assertGreater(SimpleConcurrentModel.objects.get(username='CHAR').version, 0)

    def test_creation_with_customform(self):
        url = reverse('admin:demo_simpleconcurrentmodel_add')
        res = self.app.get(url, user='sax')
        form = res.form
        username = next(nextname)
        form['username'] = username
        res = form.submit().follow()
        self.assertTrue(SimpleConcurrentModel.objects.filter(username=username).exists())
        self.assertGreater(SimpleConcurrentModel.objects.get(username=username).version, 0)

        # test no other errors are raised
        res = form.submit()
        self.assertEqual(res.status_code, 200)
        self.assertContains(res, "SimpleConcurrentModel with this Username already exists.")

    def test_standard_update(self):
        target, __ = SimpleConcurrentModel.objects.get_or_create(username='aaa')
        url = reverse('admin:demo_simpleconcurrentmodel_change', args=[target.pk])
        res = self.app.get(url, user='sax')
        target = res.context['original']
        old_version = target.version
        form = res.form
        form['username'] = 'UPDATED'
        res = form.submit().follow()
        target = SimpleConcurrentModel.objects.get(pk=target.pk)
        new_version = target.version
        self.assertGreater(new_version, old_version)

    def test_conflict(self):
        target, __ = SimpleConcurrentModel.objects.get_or_create(username='aaa')
        assert target.version
        url = reverse('admin:demo_simpleconcurrentmodel_change', args=[target.pk])
        res = self.app.get(url, user='sax')
        form = res.form

        target.save()  # create conflict here
        res = form.submit()
        self.assertIn('original', res.context)
        self.assertTrue(res.context['adminform'].form.errors,
                        res.context['adminform'].form.errors)
        self.assertIn(_('Record Modified'),
                      str(res.context['adminform'].form.errors),
                      res.context['adminform'].form.errors)

    def test_sanity_signer(self):
        target, __ = SimpleConcurrentModel.objects.get_or_create(username='aaa')
        url = reverse('admin:demo_simpleconcurrentmodel_change', args=[target.pk])
        res = self.app.get(url, user='sax')
        form = res.form
        version1 = int(str(form['version'].value).split(":")[0])
        form['version'] = VersionFieldSigner().sign(version1)
        form['date_field'] = 'esss2010-09-01'
        response = form.submit()
        self.assertIn('original', response.context)
        self.assertTrue(response.context['adminform'].form.errors,
                        response.context['adminform'].form.errors)
        form = response.context['adminform'].form
        version2 = int(str(form['version'].value()).split(":")[0])
        self.assertEqual(version1, version2)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-01-09 17:12
from __future__ import unicode_literals

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import concurrency.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='AutoIncConcurrentModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('version', concurrency.fields.AutoIncVersionField(db_column='cm_version_id', default=1, help_text='record revision number')),
                ('username', models.CharField(blank=True, max_length=30, null=True)),
                ('date_field', models.DateField(blank=True, null=True)),
            ],
            options={
                'verbose_name_plural': 'AutoIncConcurrentModel',
                'verbose_name': 'AutoIncConcurrentModel',
            },
        ),
        migrations.CreateModel(
            name='ConcreteModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('version', concurrency.fields.IntegerVersionField(db_column='cm_version_id', default=1, help_text='record revision number')),
                ('username', models.CharField(blank=True, max_length=30, null=True, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name='ConditionalVersionModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('version', concurrency.fields.ConditionalVersionField(default=1, help_text='record revision number')),
                ('field1', models.CharField(blank=True, max_length=30, null=True, unique=True)),
                ('field2', models.CharField(blank=True, max_length=30, null=True, unique=True)),
                ('field3', models.CharField(blank=True, max_length=30, null=True, unique=True)),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='DropTriggerConcurrentModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('version', concurrency.fields.TriggerVersionField(db_column='cm_version_id', default=1, help_text='record revision number')),
                ('username', models.CharField(blank=True, max_length=30, null=True)),
                ('count', models.IntegerField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name='GroupTestModel',
            fields=[
                ('group_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='auth.Group')),
                ('username', models.CharField(max_length=50, verbose_name='username')),
            ],
            bases=('auth.group',),
        ),
        migrations.CreateModel(
            name='Issue3TestModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('username', models.CharField(blank=True, max_length=30, null=True)),
                ('last_name', models.CharField(blank=True, max_length=30, null=True)),
                ('char_field', models.CharField(blank=True, max_length=30, null=True)),
                ('date_field', models.DateField(blank=True, null=True)),
                ('version', models.CharField(blank=True, default='abc', max_length=10, null=True)),
                ('revision', concurrency.fields.IntegerVersionField(db_column='cm_version_id', default=1, help_text='record revision number')),
            ],
        ),
        migrations.CreateModel(
            name='ReversionConcurrentModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('version', concurrency.fields.IntegerVersionField(db_column='cm_version_id', default=1, help_text='record revision number')),
                ('username', models.CharField(blank=True, max_length=30, null=True, unique=True)),
                ('date_field', models.DateField(blank=True, null=True)),
            ],
            options={
                'verbose_name_plural': 'Reversion-ConcurrentModels',
                'verbose_name': 'Reversion-ConcurrentModel',
            },
        ),
        migrations.CreateModel(
            name='SimpleConcurrentModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('version', concurrency.fields.IntegerVersionField(db_column='cm_version_id', default=1, help_text='record revision number')),
                ('username', models.CharField(blank=True, max_length=30, null=True, unique=True)),
                ('date_field', models.DateField(blank=True, null=True)),
            ],
            options={
                'verbose_name_plural': 'SimpleConcurrentModels',
                'verbose_name': 'SimpleConcurrentModel',
            },
        ),
        migrations.CreateModel(
            name='TriggerConcurrentModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('version', concurrency.fields.TriggerVersionField(db_column='cm_version_id', default=1, help_text='record revision number')),
                ('username', models.CharField(blank=True, max_length=30, null=True)),
                ('count', models.IntegerField(default=0)),
            ],
            options={
                'verbose_name_plural': 'TriggerConcurrentModels',
                'verbose_name': 'TriggerConcurrentModel',
            },
        ),
        migrations.CreateModel(
            name='ConcurrencyDisabledModel',
            fields=[
                ('simpleconcurrentmodel_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='demo.SimpleConcurrentModel')),
                ('dummy_char', models.CharField(blank=True, max_length=30, null=True)),
            ],
            bases=('demo.simpleconcurrentmodel',),
        ),
        migrations.CreateModel(
            name='CustomSaveModel',
            fields=[
                ('simpleconcurrentmodel_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='demo.SimpleConcurrentModel')),
                ('extra_field', models.CharField(blank=True, max_length=30, null=True, unique=True)),
            ],
            bases=('demo.simpleconcurrentmodel',),
        ),
        migrations.CreateModel(
            name='InheritedModel',
            fields=[
                ('simpleconcurrentmodel_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='demo.SimpleConcurrentModel')),
                ('extra_field', models.CharField(blank=True, max_length=30, null=True, unique=True)),
            ],
            bases=('demo.simpleconcurrentmodel',),
        ),
        migrations.CreateModel(
            name='ListEditableConcurrentModel',
            fields=[
            ],
            options={
                'verbose_name_plural': 'ListEditableConcurrentModels',
                'verbose_name': 'ListEditableConcurrentModel',
                'proxy': True,
            },
            bases=('demo.simpleconcurrentmodel',),
        ),
        migrations.CreateModel(
            name='NoActionsConcurrentModel',
            fields=[
            ],
            options={
                'verbose_name_plural': 'NoActions-ConcurrentModels',
                'verbose_name': 'NoActions-ConcurrentModel',
                'proxy': True,
            },
            bases=('demo.simpleconcurrentmodel',),
        ),
        migrations.CreateModel(
            name='ProxyModel',
            fields=[
            ],
            options={
                'verbose_name_plural': 'ProxyModels',
                'verbose_name': 'ProxyModel',
                'proxy': True,
            },
            bases=('demo.simpleconcurrentmodel',),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations

from concurrency.fields import IntegerVersionField


class Migration(migrations.Migration):
    """
    To enabe this migration you must add this code to your settings

        MIGRATION_MODULES = {
            ...
            ...
            'auth': 'tests.auth_migrations',
        }

    """
    dependencies = [
        ('auth', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='Group',
            name='version',
            field=IntegerVersionField(help_text=b'Version', default=1),

        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import logging

from django.test import TestCase

import pytest
from demo.models import SimpleConcurrentModel

from concurrency.utils import ConcurrencyTestMixin

logger = logging.getLogger(__name__)


@pytest.mark.django_db
class TestConcurrencyTestMixin(ConcurrencyTestMixin, TestCase):
    concurrency_model = SimpleConcurrentModel



# -*- coding: utf-8 -*-
from django.test.utils import override_settings

import pytest
from demo.models import AutoIncConcurrentModel, SimpleConcurrentModel
from demo.util import nextname

from concurrency.api import concurrency_disable_increment, disable_concurrency
from concurrency.exceptions import RecordModifiedError
from concurrency.utils import refetch


@pytest.mark.django_db(transaction=False)
def test_disable_concurrency_settings(settings):
    with override_settings(CONCURRENCY_ENABLED=False):
        instance1 = SimpleConcurrentModel(username=next(nextname))
        instance1.save()
        refetch(instance1).save()


@pytest.mark.django_db(transaction=False)
def test_disable_concurrency_global():
    instance1 = SimpleConcurrentModel(username=next(nextname))
    instance2 = AutoIncConcurrentModel(username=next(nextname))
    instance1.save()
    instance2.save()
    refetch(instance1).save()
    refetch(instance2).save()
    with disable_concurrency():
        instance1.save()
        instance2.save()

    copy2 = refetch(instance2)
    refetch(instance2).save()
    with pytest.raises(RecordModifiedError):
        copy2.save()


@pytest.mark.django_db(transaction=False)
def test_disable_concurrency_class(model_class=SimpleConcurrentModel):
    instance = model_class(username=next(nextname))
    instance.save()
    copy = refetch(instance)
    copy.save()
    with disable_concurrency(SimpleConcurrentModel):
        instance.save()


@pytest.mark.django_db(transaction=False)
def test_disable_concurrency_instance(model_class=SimpleConcurrentModel):
    instance1 = model_class(username=next(nextname))
    instance1.save()
    copy1 = refetch(instance1)
    copy1.save()

    instance2 = model_class(username=next(nextname))
    instance2.save()
    copy2 = refetch(instance2)
    copy2.save()

    with disable_concurrency(instance1):
        instance1.save()
        with pytest.raises(RecordModifiedError):
            instance2.save()


@pytest.mark.django_db(transaction=False)
def test_disable_increment():
    instance1 = AutoIncConcurrentModel(username=next(nextname))
    assert instance1.version == 0
    instance1.save()
    assert instance1.version == 1
    with concurrency_disable_increment(instance1):
        instance1.save()
        instance1.save()
        assert instance1.version == 1
    instance1.save()
    assert instance1.version == 2



from __future__ import absolute_import, unicode_literals

import copy
import functools
import hashlib
import logging
import time
from collections import OrderedDict
from functools import update_wrapper

from django.db.models import signals
from django.db.models.fields import Field
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from concurrency import forms
from concurrency.api import get_revision_of_object
from concurrency.config import conf
from concurrency.core import ConcurrencyOptions
from concurrency.utils import refetch

try:
    from django.apps import apps

    get_model = apps.get_model
except ImportError:
    from django.db.models.loading import get_model

try:
    from django.db.models.signals import class_prepared, post_migrate
except:
    from django.db.models.signals import class_prepared, post_syncdb as post_migrate

logger = logging.getLogger(__name__)

OFFSET = int(time.mktime((2000, 1, 1, 0, 0, 0, 0, 0, 0)))


def class_prepared_concurrency_handler(sender, **kwargs):
    if hasattr(sender, '_concurrencymeta'):
        if sender != sender._concurrencymeta.base:
            origin = getattr(sender._concurrencymeta.base, '_concurrencymeta')
            local = copy.deepcopy(origin)
            setattr(sender, '_concurrencymeta', local)

        if hasattr(sender, 'ConcurrencyMeta'):
            sender._concurrencymeta.enabled = getattr(sender.ConcurrencyMeta, 'enabled', True)
            check_fields = getattr(sender.ConcurrencyMeta, 'check_fields', None)
            ignore_fields = getattr(sender.ConcurrencyMeta, 'ignore_fields', None)
            if check_fields and ignore_fields:
                raise ValueError("Cannot set both 'check_fields' and 'ignore_fields'")

            sender._concurrencymeta.check_fields = check_fields
            sender._concurrencymeta.ignore_fields = ignore_fields
            sender._concurrencymeta.increment = getattr(sender.ConcurrencyMeta, 'increment', True)
            sender._concurrencymeta.skip = False

        if not (sender._concurrencymeta.manually):
            sender._concurrencymeta.field.wrap_model(sender)

        setattr(sender, 'get_concurrency_version', get_revision_of_object)


def post_syncdb_concurrency_handler(sender, **kwargs):
    from concurrency.triggers import create_triggers
    from django.db import connections

    databases = [alias for alias in connections]
    create_triggers(databases)


class_prepared.connect(class_prepared_concurrency_handler, dispatch_uid='class_prepared_concurrency_handler')


class TriggerRegistry(object):
    # FIXME: this is very bad. it seems required only by tests
    # see
    # https://github.com/pytest-dev/pytest-django/issues/75
    # https://code.djangoproject.com/ticket/22280#comment:20

    _fields = []

    def append(self, field):
        self._fields.append([field.model._meta.app_label, field.model.__name__])

    def __iter__(self):
        return iter([get_model(*i)._concurrencymeta.field for i in self._fields])

    def __contains__(self, field):
        target = [field.model._meta.app_label, field.model.__name__]
        return target in self._fields


_TRIGGERS = TriggerRegistry()

if not conf.MANUAL_TRIGGERS:
    post_migrate.connect(post_syncdb_concurrency_handler, dispatch_uid='post_syncdb_concurrency_handler')


class VersionField(Field):
    """ Base class """

    def __init__(self, *args, **kwargs):
        verbose_name = kwargs.get('verbose_name', None)
        name = kwargs.get('name', None)
        db_tablespace = kwargs.get('db_tablespace', None)
        db_column = kwargs.get('db_column', None)
        help_text = kwargs.get('help_text', _('record revision number'))

        super(VersionField, self).__init__(verbose_name, name,
                                           help_text=help_text,
                                           default=0,
                                           db_tablespace=db_tablespace,
                                           db_column=db_column)

    def deconstruct(self):
        name, path, args, kwargs = super(VersionField, self).deconstruct()
        kwargs['default'] = 1
        return name, path, args, kwargs

    def get_default(self):
        return 0

    def get_internal_type(self):
        return "BigIntegerField"

    def to_python(self, value):
        return int(value)

    def validate(self, value, model_instance):
        pass

    def formfield(self, **kwargs):
        kwargs['form_class'] = self.form_class
        kwargs['widget'] = forms.VersionField.widget
        return super(VersionField, self).formfield(**kwargs)

    def contribute_to_class(self, cls, name, virtual_only=False):
        super(VersionField, self).contribute_to_class(cls, name, virtual_only)
        if hasattr(cls, '_concurrencymeta') or cls._meta.abstract:
            return
        setattr(cls, '_concurrencymeta', ConcurrencyOptions())
        cls._concurrencymeta.field = self
        cls._concurrencymeta.base = cls

    def _set_version_value(self, model_instance, value):
        setattr(model_instance, self.attname, int(value))

    def pre_save(self, model_instance, add):
        if add:
            value = self._get_next_version(model_instance)
            self._set_version_value(model_instance, value)
        return getattr(model_instance, self.attname)

    @classmethod
    def wrap_model(cls, model, force=False):
        if not force and model._concurrencymeta.versioned_save:
            return
        cls._wrap_model_methods(model)
        model._concurrencymeta.versioned_save = True

    @staticmethod
    def _wrap_model_methods(model):
        old_do_update = getattr(model, '_do_update')
        setattr(model, '_do_update', model._concurrencymeta.field._wrap_do_update(old_do_update))

    def _wrap_do_update(self, func):

        def _do_update(model_instance, base_qs, using, pk_val, values, update_fields, forced_update):
            version_field = model_instance._concurrencymeta.field
            old_version = get_revision_of_object(model_instance)

            if not version_field.model._meta.abstract:
                if version_field.model is not base_qs.model:
                    return func(model_instance, base_qs, using, pk_val, values, update_fields, forced_update)

            for i, (field, _1, value) in enumerate(values):
                if field == version_field:
                    if (model_instance._concurrencymeta.increment and not
                    getattr(model_instance, '_concurrency_disable_increment', False)):
                        new_version = field._get_next_version(model_instance)
                        values[i] = (field, _1, new_version)
                        field._set_version_value(model_instance, new_version)
                    # else:
                    #     new_version = old_version
                    break

            # This provides a default if either (1) no values were provided or (2) we reached this code as part of a
            # create.  We don't need to worry about a race condition because a competing create should produce an
            # error anyway.
            updated = base_qs.filter(pk=pk_val).exists()

            # This second situation can occur because `Model.save_base` calls `Model._save_parent` without relaying
            # the `force_insert` flag that marks the process as a create.  Eventually, `Model._save_table` will call
            # this function as-if it were in the middle of an update.  The update is expected to fail because there
            # is no object to update and the caller will fall back on the create logic instead.  We need to ensure
            # the update fails (but does not raise an exception) under this circumstance by skipping the concurrency
            # logic.
            if values and updated:
                if (model_instance._concurrencymeta.enabled and
                        conf.ENABLED and
                        not getattr(model_instance, '_concurrency_disabled', False) and
                        (old_version or not conf.IGNORE_DEFAULT)):
                    filter_kwargs = {'pk': pk_val, version_field.attname: old_version}
                    updated = base_qs.filter(**filter_kwargs)._update(values) >= 1
                    if not updated:
                        version_field._set_version_value(model_instance, old_version)
                        updated = conf._callback(model_instance)
                else:
                    filter_kwargs = {'pk': pk_val}
                    updated = base_qs.filter(**filter_kwargs)._update(values) >= 1

            return updated

        return update_wrapper(_do_update, func)


class IntegerVersionField(VersionField):
    """
        Version Field that returns a "unique" version number for the record.

        The version number is produced using time.time() * 1000000, to get the benefits
        of microsecond if the system clock provides them.

    """
    form_class = forms.VersionField

    def _get_next_version(self, model_instance):
        old_value = getattr(model_instance, self.attname, 0)
        return max(int(old_value) + 1, (int(time.time() * 1000000) - OFFSET))


class AutoIncVersionField(VersionField):
    """
        Version Field increment the revision number each commit

    """
    form_class = forms.VersionField

    def _get_next_version(self, model_instance):
        return int(getattr(model_instance, self.attname, 0)) + 1


class TriggerVersionField(VersionField):
    """
        Version Field increment the revision number each commit

    """
    form_class = forms.VersionField

    def __init__(self, *args, **kwargs):
        self._trigger_name = kwargs.pop('trigger_name', None)
        self._trigger_exists = False
        super(TriggerVersionField, self).__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name, virtual_only=False):
        super(TriggerVersionField, self).contribute_to_class(cls, name)
        if not cls._meta.abstract or cls._meta.proxy:
            if self not in _TRIGGERS:
                _TRIGGERS.append(self)

    def check(self, **kwargs):
        errors = []
        model = self.model
        from django.db import router, connections
        from concurrency.triggers import factory
        from django.core.checks import Warning

        alias = router.db_for_write(model)
        connection = connections[alias]
        f = factory(connection)
        if not f.get_trigger(self):
            errors.append(
                Warning(
                    'Missed trigger for field {}'.format(self),
                    hint=None,
                    obj=None,
                    id='concurrency.W001',
                )
            )
        return errors

    @property
    def trigger_name(self):
        from concurrency.triggers import get_trigger_name

        return get_trigger_name(self)

    def _get_next_version(self, model_instance):
        # always returns the same value
        return int(getattr(model_instance, self.attname, 1))

    def pre_save(self, model_instance, add):
        # always returns the same value
        return 1

    @staticmethod
    def _increment_version_number(obj):
        old_value = get_revision_of_object(obj)
        setattr(obj, obj._concurrencymeta.field.attname, int(old_value) + 1)

    @staticmethod
    def _wrap_model_methods(model):
        super(TriggerVersionField, TriggerVersionField)._wrap_model_methods(model)
        old_save = getattr(model, 'save')
        setattr(model, 'save', model._concurrencymeta.field._wrap_save(old_save))

    @staticmethod
    def _wrap_save(func):
        def inner(self, force_insert=False, force_update=False, using=None, **kwargs):
            reload = kwargs.pop('refetch', False)
            ret = func(self, force_insert, force_update, using, **kwargs)
            TriggerVersionField._increment_version_number(self)
            if reload:
                ret = refetch(self)
                setattr(self,
                        self._concurrencymeta.field.attname,
                        get_revision_of_object(ret))
            return ret

        return update_wrapper(inner, func)


def filter_fields(instance, field):
    if not field.concrete:
        # reverse relation
        return False
    if field.is_relation and field.related_model is None:
        # generic foreignkeys
        return False
    if field.many_to_many and instance.pk is None:
        # can't load remote object yet
        return False
    return True


class ConditionalVersionField(AutoIncVersionField):
    def contribute_to_class(self, cls, name, virtual_only=False):
        super(ConditionalVersionField, self).contribute_to_class(cls, name, virtual_only)
        signals.post_init.connect(self._load_model,
                                  sender=cls, weak=False)

        signals.post_save.connect(self._save_model,
                                  sender=cls, weak=False)

    def _load_model(self, *args, **kwargs):
        instance = kwargs['instance']
        instance._concurrencymeta.initial = self._get_hash(instance)

    def _save_model(self, *args, **kwargs):
        instance = kwargs['instance']
        instance._concurrencymeta.initial = self._get_hash(instance)

    def _get_hash(self, instance):
        values = OrderedDict()
        opts = instance._meta
        check_fields = instance._concurrencymeta.check_fields
        ignore_fields = instance._concurrencymeta.ignore_fields

        filter_ = functools.partial(filter_fields, instance)
        if check_fields is None and ignore_fields is None:
            fields = sorted([f.name for f in filter(filter_, instance._meta.get_fields())])
        elif check_fields is None:
            fields = sorted([f.name for f in filter(filter_, instance._meta.get_fields())
                             if f.name not in ignore_fields])
        else:
            fields = instance._concurrencymeta.check_fields

        for field_name in fields:
            # do not use getattr here. we do not need extra sql to retrieve
            # FK. the raw value of the FK is enough
            values[field_name] = opts.get_field(field_name).value_from_object(instance)
        return hashlib.sha1(force_text(values).encode('utf-8')).hexdigest()

    def _get_next_version(self, model_instance):
        if not model_instance.pk:
            return int(getattr(model_instance, self.attname) + 1)

        old = model_instance._concurrencymeta.initial
        new = self._get_hash(model_instance)
        if old != new:
            return int(getattr(model_instance, self.attname, 0) + 1)

        return int(getattr(model_instance, self.attname, 0))



# -*- coding: utf-8 -*-
from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import BaseCommand
from django.db import connections

from concurrency.compat import atomic
from concurrency.triggers import create_triggers, drop_triggers, get_triggers


class Command(BaseCommand):
    args = ''
    help = 'register Report classes and create one ReportConfiguration per each'

    requires_system_checks = False

    def add_arguments(self, parser):
        """
        Entry point for subclassed commands to add custom arguments.
        """
        subparsers = parser.add_subparsers(help='sub-command help',
                                           dest='command')

        subparsers.add_parser('list',
                              cmd=parser,
                              help="command_a help")
        subparsers.add_parser('drop',
                              cmd=parser,
                              help="command_a help")
        subparsers.add_parser('create',
                              cmd=parser,
                              help="command_a help")

        parser.add_argument('-d', '--database',
                            action='store',
                            dest='database',
                            default=None,
                            help='limit to this database')

        parser.add_argument('-t', '--trigger',
                            action='store',
                            dest='trigger',
                            default=None,
                            help='limit to this trigger name')

    def _list(self, databases):
        for alias, triggers in get_triggers(databases).items():
            self.stdout.write("Database: {}".format(alias))
            for trigger in triggers:
                self.stdout.write("       {}".format(trigger))
        self.stdout.write('')

    def handle(self, *args, **options):

        cmd = options['command']
        database = options['database']
        if database is None:
            databases = [alias for alias in connections]
        else:
            databases = [database]

        with atomic():
            try:
                if cmd == 'list':
                    self._list(databases)
                elif cmd == 'create':
                    for alias, triggers in create_triggers(databases).items():
                        self.stdout.write("Database: {}".format(alias))
                        for trigger in triggers:
                            self.stdout.write("    Created {0[2]}  for {0[1]}".format(trigger))
                    self.stdout.write('')
                elif cmd == 'drop':
                    for alias, triggers in drop_triggers(*databases).items():
                        self.stdout.write("Database: {}".format(alias))
                        for trigger in triggers:
                            self.stdout.write("    Dropped   {0[2]}".format(trigger))
                    self.stdout.write('')
                else:
                    raise Exception()
            except ImproperlyConfigured as e:
                self.stdout.write(self.style.ERROR(e))



import pytest
from demo.models import (
    AutoIncConcurrentModel, ConcreteModel, CustomSaveModel, InheritedModel, ProxyModel,
    SimpleConcurrentModel
)
from demo.util import nextname, unique_id, with_models, with_std_models

from concurrency.exceptions import RecordModifiedError
from concurrency.utils import refetch


@pytest.mark.django_db
@with_std_models
def test_get_or_create(model_class):
    instance, __ = model_class.objects.get_or_create(pk=next(unique_id))
    assert instance.get_concurrency_version()
    instance.save()


@pytest.mark.django_db
@with_std_models
def test_get_or_create_with_pk(model_class):
    instance, __ = model_class.objects.get_or_create(pk=next(unique_id))
    assert instance.get_concurrency_version()
    instance.save()
    copy = refetch(instance)
    copy.save()
    with pytest.raises(RecordModifiedError):
        instance.save()
    assert copy.get_concurrency_version() > instance.get_concurrency_version()


@pytest.mark.django_db(transaction=False)
def test_create(model_class=SimpleConcurrentModel):
    instance = model_class.objects.create(pk=next(unique_id))
    assert instance.get_concurrency_version()


@pytest.mark.django_db
@with_models(SimpleConcurrentModel, AutoIncConcurrentModel,
             InheritedModel, CustomSaveModel,
             ConcreteModel, ProxyModel)
def test_update(model_class):
    # Manager.update() does not change version number
    instance = model_class.objects.create(pk=next(unique_id), username=next(nextname).lower())
    field_value = instance.username
    model_class.objects.filter(pk=instance.pk).update(username=instance.username.upper())

    instance2 = refetch(instance)
    assert instance2.username == field_value.upper()
    assert instance2.get_concurrency_version() == instance.get_concurrency_version()



from django.contrib.auth.models import Group

import pytest
from demo.models import SimpleConcurrentModel
from demo.util import nextgroup, nextname

from concurrency.api import (
    apply_concurrency_check, get_revision_of_object, get_version, is_changed
)
from concurrency.exceptions import RecordModifiedError
from concurrency.fields import IntegerVersionField
from concurrency.utils import refetch


@pytest.mark.django_db(transaction=False)
@pytest.mark.skipIf('os.environ["DBENGINE"]=="pg"')
def test_get_revision_of_object(model_class=SimpleConcurrentModel):
    instance = model_class(username=next(nextname))
    instance.save()
    assert get_revision_of_object(instance) == instance.version


@pytest.mark.django_db
def test_is_changed(model_class=SimpleConcurrentModel):
    instance = model_class(username=next(nextname))
    instance.save()
    copy = refetch(instance)
    copy.save()
    assert is_changed(instance)


@pytest.mark.django_db
def test_get_version(model_class=SimpleConcurrentModel):
    instance = model_class(username=next(nextname))
    instance.save()
    copy = refetch(instance)
    copy.save()
    instance = get_version(instance, copy.version)
    assert instance.get_concurrency_version() == copy.get_concurrency_version()


@pytest.mark.django_db(transaction=False)
def test_apply_concurrency_check():
    apply_concurrency_check(Group, 'version', IntegerVersionField)

    instance, __ = Group.objects.get_or_create(name=next(nextgroup))
    instance.save()

    copy = refetch(instance)
    copy.save()

    with pytest.raises(RecordModifiedError):
        instance.save()



from __future__ import unicode_literals

from django.contrib import admin
from django.contrib.admin.sites import NotRegistered

from demo.models import *  # noqa
from demo.models import (
    ListEditableConcurrentModel, NoActionsConcurrentModel, ReversionConcurrentModel
)

from concurrency.admin import ConcurrentModelAdmin
from concurrency.api import disable_concurrency

try:
    from reversion import VersionAdmin
except ImportError:
    class VersionAdmin(object):
        pass


class ListEditableModelAdmin(ConcurrentModelAdmin):
    list_display = ('__unicode__', 'version', 'username')
    list_editable = ('username',)
    ordering = ('id',)


class NoActionsModelAdmin(ConcurrentModelAdmin):
    list_display = ('__unicode__', 'version', 'username')
    list_editable = ('username',)
    ordering = ('id',)
    actions = None


class ReversionConcurrentModelAdmin(VersionAdmin, ConcurrentModelAdmin):
    list_display = ('__unicode__', 'version', 'username')
    list_editable = ('username',)
    ordering = ('id',)
    actions = None

    @disable_concurrency()
    def recover_view(self, request, version_id, extra_context=None):
        return super(ReversionConcurrentModelAdmin, self).recover_view(request,
                                                                       version_id,
                                                                       extra_context)


class ActionsModelAdmin(ConcurrentModelAdmin):
    list_display = ('__unicode__', 'version', 'username')
    actions = ['dummy_action']
    ordering = ('id',)

    def dummy_action(self, request, queryset):
        for el in queryset:
            el.username = '**action_update**'
            el.save()


def admin_register(model, modeladmin=ConcurrentModelAdmin):
    try:
        admin.site.unregister(model)
    except NotRegistered:  # pragma: no cover
        pass
    admin.site.register(model, modeladmin)


def admin_register_models():
    admin_register(SimpleConcurrentModel, ActionsModelAdmin)
    admin_register(ProxyModel, ListEditableModelAdmin)
    admin_register(InheritedModel, ActionsModelAdmin)
    admin_register(NoActionsConcurrentModel, NoActionsModelAdmin)
    admin_register(ListEditableConcurrentModel, ListEditableModelAdmin)
    admin_register(ReversionConcurrentModel, ReversionConcurrentModelAdmin)


admin_register_models()



from __future__ import unicode_literals

import django
from django.contrib.auth.models import Group, User
from django.test import TransactionTestCase
from django.utils import timezone

from demo.admin import admin_register_models
from django_webtest import WebTestMixin

from concurrency.api import apply_concurrency_check
from concurrency.fields import IntegerVersionField

SENTINEL = '**concurrent_update**'


apply_concurrency_check(Group, 'version', IntegerVersionField)

DJANGO_TRUNK = django.VERSION[:2] >= (1, 8)


class AdminTestCase(WebTestMixin, TransactionTestCase):
    urls = 'demo.urls'

    def setUp(self):
        super(AdminTestCase, self).setUp()

        self.user, __ = User.objects.get_or_create(is_superuser=True,
                                                   is_staff=True,
                                                   is_active=True,
                                                   last_login=timezone.now(),
                                                   email='sax@example.com',
                                                   username='sax')
        admin_register_models()


# class DjangoAdminTestCase(TransactionTestCase):
#     urls = 'concurrency.tests.urls'
#     MIDDLEWARE_CLASSES = global_settings.MIDDLEWARE_CLASSES
#     AUTHENTICATION_BACKENDS = global_settings.AUTHENTICATION_BACKENDS
#
#     def setUp(self):
#         super(DjangoAdminTestCase, self).setUp()
#         self.sett = self.settings(
#             #INSTALLED_APPS=INSTALLED_APPS,
#             MIDDLEWARE_CLASSES=self.MIDDLEWARE_CLASSES,
#             AUTHENTICATION_BACKENDS=self.AUTHENTICATION_BACKENDS,
#             PASSWORD_HASHERS=('django.contrib.auth.hashers.MD5PasswordHasher',),  # fastest hasher
#             STATIC_URL='/static/',
#             SOUTH_TESTS_MIGRATE=False,
#             TEMPLATE_DIRS=(os.path.join(os.path.dirname(__file__), 'templates'),))
#         self.sett.enable()
#         django.core.management._commands = None  # reset commands cache
#         django.core.management.call_command('syncdb', verbosity=0)
#
#         # admin_register(TestModel0)
#         # admin_register(TestModel1, TestModel1Admin)
#
#         self.user, __ = User.objects.get_or_create(username='sax',
#                                                    is_active=True,
#                                                    is_staff=True,
#                                                    is_superuser=True)
#         self.user.set_password('123')
#         self.user.save()
#         self.client.login(username=self.user.username, password='123')
#         # self.target, __ = TestModel0.objects.get_or_create(username='aaa')
#         # self.target1, __ = TestModel1.objects.get_or_create(username='bbb')
#
#     def tearDown(self):
#         super(DjangoAdminTestCase, self).tearDown()
#         self.sett.disable()
#         # admin_unregister(TestModel0, TestModel1)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import logging

import pytest
from demo.models import SimpleConcurrentModel

from concurrency.templatetags.concurrency import identity, is_version, version

logger = logging.getLogger(__name__)


@pytest.fixture
def obj():
    return SimpleConcurrentModel.objects.create()


@pytest.mark.django_db
def test_identity(obj):
    assert identity(obj).split(',') == [str(obj.pk), str(obj.version)]


@pytest.mark.django_db
def test_version(obj):
    assert version(obj) == obj.version


@pytest.mark.django_db
def test_is_version(obj):
    assert is_version(obj._concurrencymeta.field)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import logging

import django

import pytest
from demo.models import TriggerConcurrentModel

logger = logging.getLogger(__name__)


@pytest.fixture
def obj():
    return TriggerConcurrentModel.objects.create()


@pytest.mark.skipif(django.VERSION[:2] < (1, 7),
                    reason="Skip if django< 1.7")
@pytest.mark.django_db
def test_check(obj, monkeypatch):
    from django.core.checks import Warning
    monkeypatch.setattr(obj._concurrencymeta.field, '_trigger_name', 'test')

    assert isinstance(obj._concurrencymeta.field.check()[0], Warning)



from django.test import TransactionTestCase

from demo.models import ConcurrencyDisabledModel, SimpleConcurrentModel

from concurrency.exceptions import RecordModifiedError


class TestCustomConcurrencyMeta(TransactionTestCase):
    concurrency_model = ConcurrencyDisabledModel
    concurrency_kwargs = {'username': 'test'}

    def setUp(self):
        super(TestCustomConcurrencyMeta, self).setUp()
        self.TARGET = self._get_concurrency_target()

    def _get_concurrency_target(self, **kwargs):
        # WARNING this method must be idempotent. ie must returns
        # always a fresh copy of the record
        args = dict(self.concurrency_kwargs)
        args.update(kwargs)
        return self.concurrency_model.objects.get_or_create(**args)[0]

    def test_enabled(self):
        assert not self.TARGET._concurrencymeta.enabled

    def test_meta_inheritance(self):
        # TestModelWithCustomOptions extends ConcurrentModel
        # but we disabled concurrency only in TestModelWithCustomOptions
        import concurrency.api as api
        concurrency_enabled1 = SimpleConcurrentModel.objects.get_or_create(**{'username': 'test'})[0]
        concurrency_enabled2 = SimpleConcurrentModel.objects.get_or_create(**{'username': 'test'})[0]
        v1 = api.get_revision_of_object(concurrency_enabled1)
        v2 = api.get_revision_of_object(concurrency_enabled2)
        assert v1 == v2, "got same row with different version (%s/%s)" % (v1, v2)
        concurrency_enabled1.save()
        assert concurrency_enabled1.pk is not None  # sanity check
        self.assertRaises(RecordModifiedError, concurrency_enabled2.save)

        concurrency_disabled1 = ConcurrencyDisabledModel.objects.get_or_create(**{'username': 'test'})[0]
        concurrency_disabled2 = ConcurrencyDisabledModel.objects.get_or_create(**{'username': 'test'})[0]
        v1 = api.get_revision_of_object(concurrency_disabled1)
        v2 = api.get_revision_of_object(concurrency_disabled2)
        assert v1 == v2, "got same row with different version (%s/%s)" % (v1, v2)
        concurrency_disabled1.save()
        assert concurrency_disabled1.pk is not None  # sanity check
        v1 = api.get_revision_of_object(concurrency_disabled1)
        v2 = api.get_revision_of_object(concurrency_disabled2)
        assert v1 != v2



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib.auth.models import Group, User
from django.db import models

from concurrency.fields import (
    AutoIncVersionField, ConditionalVersionField, IntegerVersionField,
    TriggerVersionField
)

__all__ = ['SimpleConcurrentModel', 'AutoIncConcurrentModel',
           'ProxyModel', 'InheritedModel', 'CustomSaveModel',
           'ConcreteModel', 'TriggerConcurrentModel',
           'ConditionalVersionModelWithoutMeta',
           'Anything',
           ]


class SimpleConcurrentModel(models.Model):
    version = IntegerVersionField(db_column='cm_version_id')
    username = models.CharField(max_length=30, blank=True, null=True, unique=True)
    date_field = models.DateField(blank=True, null=True)

    class Meta:
        app_label = 'demo'
        verbose_name = "SimpleConcurrentModel"
        verbose_name_plural = "SimpleConcurrentModels"

    def __str__(self):
        return "{0} #{1}".format(self.__class__.__name__, self.pk)

    def __unicode__(self):
        return "{0} #{1}".format(self.__class__.__name__, self.pk)


class AutoIncConcurrentModel(models.Model):
    version = AutoIncVersionField(db_column='cm_version_id')
    username = models.CharField(max_length=30, blank=True, null=True)
    date_field = models.DateField(blank=True, null=True)

    class Meta:
        app_label = 'demo'
        verbose_name = "AutoIncConcurrentModel"
        verbose_name_plural = "AutoIncConcurrentModel"

    def __unicode__(self):
        return "{0.__class__.__name__} #{0.pk}".format(self)


class TriggerConcurrentModel(models.Model):
    version = TriggerVersionField(db_column='cm_version_id')
    username = models.CharField(max_length=30, blank=True, null=True)
    count = models.IntegerField(default=0)

    class Meta:
        app_label = 'demo'
        verbose_name = "TriggerConcurrentModel"
        verbose_name_plural = "TriggerConcurrentModels"

    def __unicode__(self):
        return "{0.__class__.__name__} #{0.pk}".format(self)


class DropTriggerConcurrentModel(models.Model):
    # only used to test drop triggers
    version = TriggerVersionField(db_column='cm_version_id')
    username = models.CharField(max_length=30, blank=True, null=True)
    count = models.IntegerField(default=0)

    class Meta:
        app_label = 'demo'


class ProxyModel(SimpleConcurrentModel):
    class Meta:
        app_label = 'demo'
        proxy = True
        verbose_name = "ProxyModel"
        verbose_name_plural = "ProxyModels"


class InheritedModel(SimpleConcurrentModel):
    extra_field = models.CharField(max_length=30, blank=True, null=True, unique=True)

    class Meta:
        app_label = 'demo'


class CustomSaveModel(SimpleConcurrentModel):
    extra_field = models.CharField(max_length=30, blank=True, null=True, unique=True)

    def save(self, *args, **kwargs):
        super(CustomSaveModel, self).save(*args, **kwargs)

    class Meta:
        app_label = 'demo'


class AbstractModel(models.Model):
    version = IntegerVersionField(db_column='cm_version_id')
    username = models.CharField(max_length=30, blank=True, null=True, unique=True)

    class Meta:
        app_label = 'demo'
        abstract = True


class ConcreteModel(AbstractModel):
    pass

    class Meta:
        app_label = 'demo'


# class TestCustomUser(User):
# version = IntegerVersionField(db_column='cm_version_id')
#
#     class Meta:
#         app_label = 'demo'
#
#     def __unicode__(self):
#         return "{0.__class__.__name__} #{0.pk}".format(self)


class GroupTestModel(Group):
    # HACK: this field is here because all tests relies on that
    # and we need a 'fresh' model to check for on-the-fly addition
    # of version field.  (added in tests 0.3.0)

    username = models.CharField('username', max_length=50)

    class Meta:
        app_label = 'demo'


# class TestModelGroupWithCustomSave(TestModelGroup):
#     class Meta:
#         app_label = 'demo'
#
#     def save(self, *args, **kwargs):
#         super(TestModelGroupWithCustomSave, self).save(*args, **kwargs)
#         return 222


class Issue3TestModel(models.Model):
    username = models.CharField(max_length=30, blank=True, null=True)
    last_name = models.CharField(max_length=30, blank=True, null=True)
    char_field = models.CharField(max_length=30, blank=True, null=True)
    date_field = models.DateField(blank=True, null=True)

    version = models.CharField(default='abc', max_length=10, blank=True, null=True)
    revision = IntegerVersionField(db_column='cm_version_id')

    class Meta:
        app_label = 'demo'


class ListEditableConcurrentModel(SimpleConcurrentModel):
    """ Proxy model used by admin related test.
    This allow to use multiple ModelAdmin configuration with the same 'real' model
    """

    class Meta:
        app_label = 'demo'
        proxy = True
        verbose_name = "ListEditableConcurrentModel"
        verbose_name_plural = "ListEditableConcurrentModels"


class NoActionsConcurrentModel(SimpleConcurrentModel):
    """ Proxy model used by admin related test.
    This allow to use multiple ModelAdmin configuration with the same 'real' model
    """

    class Meta:
        app_label = 'demo'
        proxy = True
        verbose_name = "NoActions-ConcurrentModel"
        verbose_name_plural = "NoActions-ConcurrentModels"


class ReversionConcurrentModel(models.Model):
    version = IntegerVersionField(db_column='cm_version_id')
    username = models.CharField(max_length=30, blank=True, null=True, unique=True)
    date_field = models.DateField(blank=True, null=True)

    class Meta:
        app_label = 'demo'
        verbose_name = "Reversion-ConcurrentModel"
        verbose_name_plural = "Reversion-ConcurrentModels"

    def __unicode__(self):
        return "{0.__class__.__name__} #{0.pk}".format(self)


class ConcurrencyDisabledModel(SimpleConcurrentModel):
    dummy_char = models.CharField(max_length=30, blank=True, null=True)

    class Meta:
        app_label = 'demo'

    class ConcurrencyMeta:
        enabled = False


class ConditionalVersionModel(models.Model):
    version = ConditionalVersionField()
    field1 = models.CharField(max_length=30, blank=True, null=True, unique=True)
    field2 = models.CharField(max_length=30, blank=True, null=True, unique=True)
    field3 = models.CharField(max_length=30, blank=True, null=True, unique=True)
    user = models.ForeignKey(User, null=True)

    class Meta:
        app_label = 'demo'

    class ConcurrencyMeta:
        check_fields = ['field1', 'field2', 'user']


class Anything(models.Model):
    """
    Will create a ManyToOneRel automatic field on
    ConditionalVersionModelWithoutMeta instances.
    """
    name = models.CharField(max_length=10)
    a_relation = models.ForeignKey('demo.ConditionalVersionModelWithoutMeta')


class ConditionalVersionModelWithoutMeta(models.Model):
    """
    This model doesn't have ConcurrencyMeta defined.
    """
    version = ConditionalVersionField()
    field1 = models.CharField(max_length=30, blank=True, null=True, unique=True)
    field2 = models.CharField(max_length=30, blank=True, null=True, unique=True)
    field3 = models.CharField(max_length=30, blank=True, null=True, unique=True)
    user = models.ForeignKey(User, null=True)
    anythings = models.ManyToManyField(Anything)

    class Meta:
        app_label = 'demo'



import itertools
from contextlib import contextmanager
from functools import partial, update_wrapper
from itertools import count

from django import db

import pytest
from demo.models import (
    AutoIncConcurrentModel, ConcreteModel, CustomSaveModel, InheritedModel, ProxyModel,
    SimpleConcurrentModel, TriggerConcurrentModel
)

from concurrency.config import conf


def sequence(prefix):
    infinite = itertools.count()
    while 1:
        yield "{0}-{1}".format(prefix, next(infinite))


nextname = sequence('username')
nextgroup = sequence('group')
unique_id = count(1)


def override_conf(**kwargs):
    for key, new_value in kwargs.items():
        setattr(conf, key, new_value)


def clone_instance(model_instance):
    """
        returns a copy of the passed instance.

        .. warning: All fields are copied, even primary key

    :param instance: :py:class:`django.db.models.Model` instance
    :return: :py:class:`django.db.models.Model` instance
    """

    fieldnames = [fld.name for fld in model_instance._meta.fields]

    new_kwargs = {name: getattr(model_instance, name) for name in fieldnames}
    return model_instance.__class__(**new_kwargs)


def with_models(*models, **kwargs):
    ignore = kwargs.pop('ignore', [])
    if ignore:
        models = filter(models, lambda x: x not in ignore)

    ids = [m.__name__ for m in models]

    return pytest.mark.parametrize(('model_class,'),
                                   models,
                                   False,
                                   ids,
                                   None)


MODEL_CLASSES = [SimpleConcurrentModel, AutoIncConcurrentModel,
                 InheritedModel, CustomSaveModel,
                 ConcreteModel, ProxyModel, TriggerConcurrentModel]

with_std_models = partial(with_models, SimpleConcurrentModel, AutoIncConcurrentModel,
                          InheritedModel, CustomSaveModel,
                          ConcreteModel, ProxyModel)()
with_all_models = partial(with_models, *MODEL_CLASSES)()

# with_all_models = partial(models_parametrize, ConcreteModel)()

DELETE_ATTRIBUTE = object()


@pytest.fixture(params=MODEL_CLASSES)
def concurrent_model(request):
    return request.param


@contextmanager
def attributes(*values):
    """
        context manager to temporary set/delete object's attributes
    :param values: tulples of (target, name, value)
    Es.


    with attributes((django.contrib.admin.ModelAdmin, 'list_per_page', 200)):
        ...

    with attributes((django.contrib.admin.ModelAdmin, 'list_per_page', DELETE_ATTRIBUTE)):
        ...

    """

    def _set(target, name, value):
        if value is DELETE_ATTRIBUTE:
            delattr(target, name)
        else:
            setattr(target, name, value)

    backups = []

    for target, name, value in values:
        if hasattr(target, name):
            backups.append((target, name, getattr(target, name)))
        else:
            backups.append((target, name, getattr(target, name, DELETE_ATTRIBUTE)))
        _set(target, name, value)
    yield

    for target, name, value in backups:
        _set(target, name, value)


def concurrently(times=1):
    # from: http://www.caktusgroup.com/blog/2009/05/26/testing-django-views-for-concurrency-issues/
    """
Add this decorator to small pieces of code that you want to test
concurrently to make sure they don't raise exceptions when run at the
same time. E.g., some Django views that do a SELECT and then a subsequent
INSERT might fail when the INSERT assumes that the data has not changed
since the SELECT.
"""

    def concurrently_decorator(test_func):
        def wrapper(*args, **kwargs):
            exceptions = []
            import threading

            def call_test_func():
                try:
                    test_func(*args, **kwargs)
                except Exception as e:
                    exceptions.append(e)
                    raise
                finally:
                    db.connection.close()

            threads = []
            for i in range(times):
                threads.append(threading.Thread(target=call_test_func))
            for t in threads:
                t.start()
            for t in threads:
                t.join()
            if exceptions:
                raise Exception(
                    'test_concurrently intercepted %s exceptions: %s' %
                    (len(exceptions), exceptions))

        return update_wrapper(wrapper, test_func)

    return concurrently_decorator



import pytest
from django.test import override_settings

from demo.util import concurrent_model, unique_id, with_all_models, with_std_models

from concurrency.core import _set_version
from concurrency.exceptions import RecordModifiedError
from concurrency.utils import refetch

pytest.mark.django_db(transaction=False)


@with_all_models
@pytest.mark.django_db
def test_standard_save(model_class):
    instance = model_class(username=concurrent_model.__name__)
    instance.save()
    assert instance.get_concurrency_version() > 0
    instance = refetch(instance)
    assert instance.get_concurrency_version() > 0


@pytest.mark.django_db(transaction=False)
@with_std_models
def test_conflict(model_class):
    id = next(unique_id)
    instance = model_class.objects.get_or_create(pk=id)[0]
    instance.save()

    copy = refetch(instance)
    copy.save()

    with pytest.raises(RecordModifiedError):
        instance.save()
    assert copy.get_concurrency_version() > instance.get_concurrency_version()


@pytest.mark.django_db(transaction=True)
@with_std_models
def test_do_not_check_if_no_version(model_class):
    id = next(unique_id)
    instance, __ = model_class.objects.get_or_create(pk=id)
    instance.save()

    copy = refetch(instance)
    copy.save()

    with pytest.raises(RecordModifiedError):
        _set_version(instance, 1)
        instance.version = 1
        instance.save()

    _set_version(instance, 0)
    instance.save()
    assert instance.get_concurrency_version() > 0
    assert instance.get_concurrency_version() != copy.get_concurrency_version()


@pytest.mark.django_db(transaction=True)
@with_std_models
def test_conflict_no_version_and_no_skip_flag(model_class):
    """When IGNORE_DEFAULT is disabled, attempting to update a record with a default version number should fail."""
    with override_settings(CONCURRENCY_IGNORE_DEFAULT=False):
        id = next(unique_id)
        instance, __ = model_class.objects.get_or_create(pk=id)
        instance.save()

        copy = refetch(instance)
        copy.version = 0

        with pytest.raises(RecordModifiedError):
            copy.save()


@with_std_models
@pytest.mark.django_db(transaction=False)
def test_update_fields(model_class):
    """
    Calling save with update_fields not containing version doesn't update
    the version.
    """

    instance = model_class.objects.create(username='abc')
    copy = refetch(instance)

    # do not update version
    instance.save(update_fields=['username'])

    # copy can be saved
    copy.username = 'def'
    copy.save()
    assert refetch(instance).username, 'def'
    assert refetch(instance).version == copy.version


@with_std_models
@pytest.mark.django_db(transaction=False)
def test_update_fields_still_checks(model_class):
    """
    Excluding the VersionField from update_fields should still check
    for conflicts.
    """
    instance = model_class.objects.create(username='abc')
    copy = refetch(instance)
    instance.save()
    copy.name = 'def'

    with pytest.raises(RecordModifiedError):
        copy.save(update_fields=['username'])



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from collections import defaultdict

from django.db import connections, router
from django.db.utils import DatabaseError

from concurrency.fields import _TRIGGERS  # noqa


def get_trigger_name(field):
    """

    :param field: Field instance
    :return: unicode
    """
    if field._trigger_name:
        name = field._trigger_name
    else:
        name = '{1.db_table}_{0.name}'.format(field, field.model._meta)
    return 'concurrency_{}'.format(name)


def get_triggers(databases=None):
    if databases is None:
        databases = [alias for alias in connections]

    ret = {}
    for alias in databases:
        connection = connections[alias]
        f = factory(connection)
        r = f.get_list()
        ret[alias] = r
    return ret


def drop_triggers(*databases):
    global _TRIGGERS
    ret = defaultdict(lambda: [])
    for field in set(_TRIGGERS):
        model = field.model
        alias = router.db_for_write(model)
        if alias in databases:
            connection = connections[alias]
            f = factory(connection)
            f.drop(field)
            field._trigger_exists = False
            ret[alias].append([model, field, field.trigger_name])
    return ret


def create_triggers(databases):
    global _TRIGGERS
    ret = defaultdict(lambda: [])

    for field in set(_TRIGGERS):
        model = field.model
        alias = router.db_for_write(model)
        if alias in databases:
            if not field._trigger_exists:
                field._trigger_exists = True
                connection = connections[alias]
                f = factory(connection)
                f.create(field)
                ret[alias].append([model, field, field.trigger_name])
    # _TRIGGERS = []
    return ret


class TriggerFactory(object):
    update_clause = ""
    drop_clause = ""
    list_clause = ""

    def __init__(self, connection):
        self.connection = connection

    def get_trigger(self, field):
        if field.trigger_name in self.get_list():
            return field.trigger_name
        return None

    def create(self, field):
        if field.trigger_name not in self.get_list():
            stm = self.update_clause.format(trigger_name=field.trigger_name,
                                            opts=field.model._meta,
                                            field=field)
            try:
                self.connection.cursor().execute(stm)
            except BaseException as exc:  # pragma: no cover
                raise DatabaseError("""Error executing:
{1}
{0}""".format(exc, stm))
        field._trigger_exists = True

    def drop(self, field):
        opts = field.model._meta
        ret = []
        stm = self.drop_clause.format(trigger_name=field.trigger_name,
                                      opts=opts,
                                      field=field)
        self.connection.cursor().execute(stm)
        ret.append(field.trigger_name)
        return ret

    def _list(self):
        cursor = self.connection.cursor()
        cursor.execute(self.list_clause)
        return cursor.fetchall()

    def get_list(self):
        return sorted([m[0] for m in self._list()])


class Sqlite3(TriggerFactory):
    drop_clause = """DROP TRIGGER IF EXISTS {trigger_name};"""

    update_clause = """CREATE TRIGGER {trigger_name}
AFTER UPDATE ON {opts.db_table}
BEGIN UPDATE {opts.db_table} SET {field.column} = {field.column}+1 WHERE {opts.pk.column} = NEW.{opts.pk.column};
END;"""

    list_clause = "select name from sqlite_master where type='trigger';"


class PostgreSQL(TriggerFactory):
    drop_clause = r"""DROP TRIGGER IF EXISTS {trigger_name} ON {opts.db_table};"""

    update_clause = r"""CREATE OR REPLACE FUNCTION func_{trigger_name}()
    RETURNS TRIGGER as
    '
    BEGIN
       NEW.{field.column} = OLD.{field.column} +1;
        RETURN NEW;
    END;
    ' language 'plpgsql';

CREATE TRIGGER {trigger_name} BEFORE UPDATE
    ON {opts.db_table} FOR EACH ROW
    EXECUTE PROCEDURE func_{trigger_name}();
    """

    list_clause = "select * from pg_trigger where tgname LIKE 'concurrency_%%'; "

    def get_list(self):
        return sorted([m[1] for m in self._list()])


class MySQL(TriggerFactory):
    drop_clause = """DROP TRIGGER IF EXISTS {trigger_name};"""

    update_clause = """
CREATE TRIGGER {trigger_name} BEFORE UPDATE ON {opts.db_table}
FOR EACH ROW SET NEW.{field.column} = OLD.{field.column}+1;
"""

    list_clause = "SHOW TRIGGERS"


def factory(conn):
    try:
        return {'postgresql': PostgreSQL,
                'mysql': MySQL,
                'sqlite3': Sqlite3,
                'sqlite': Sqlite3,
                }[conn.vendor](conn)
    except KeyError:
        raise ValueError('{} is not supported by TriggerVersionField'.format(conn))



from __future__ import absolute_import, unicode_literals

try:
    from django.apps import AppConfig

    class ConcurrencyConfig(AppConfig):
        name = 'concurrency'
        verbose = 'Django Concurrency'

except ImportError:  # pragma: no cover
    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

try:
    from django.template.base import TemplateDoesNotExist
except ImportError:
    from django.template.exceptions import TemplateDoesNotExist  # noqa - django 1.9

try:
    from django.db.transaction import atomic
except ImportError:
    from django.db.transaction import commit_on_success as atomic  # noqa



import sys

from django import db
from django.db import transaction

import pytest
from demo.models import TriggerConcurrentModel
from demo.util import concurrently

from concurrency.exceptions import RecordModifiedError
from concurrency.utils import refetch
from conftest import skippypy


@skippypy
@pytest.mark.django_db(transaction=True)
def test_threads():
    if db.connection.vendor == 'sqlite':
        pytest.skip("in-memory sqlite db can't be used between threads")

    obj = TriggerConcurrentModel.objects.create()
    transaction.commit()

    @concurrently(25)
    def run():
        for i in range(5):
            while True:
                x = refetch(obj)
                transaction.commit()
                x.count += 1
                try:
                    x.save()
                    transaction.commit()
                except RecordModifiedError:
                    # retry
                    pass
                else:
                    break

    run()
    assert refetch(obj).count == 5 * 25



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from django.contrib.admin.models import LogEntry
from django.contrib.admin.sites import site
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.utils.encoding import force_text

import pytest
from demo.base import SENTINEL, AdminTestCase
from demo.models import ListEditableConcurrentModel
from demo.util import attributes, unique_id

from concurrency.config import (
    CONCURRENCY_LIST_EDITABLE_POLICY_ABORT_ALL, CONCURRENCY_LIST_EDITABLE_POLICY_SILENT
)
from concurrency.exceptions import RecordModifiedError


@pytest.mark.xfail("django.VERSION[:2] == (1, 10)", strict=True)
class TestListEditable(AdminTestCase):
    TARGET = ListEditableConcurrentModel

    def _create_conflict(self, pk):
        u = self.TARGET.objects.get(pk=pk)
        u.username = SENTINEL
        u.save()

    def test_normal_add(self):
        res = self.app.get('/admin/', user='sax')

        res = res.click(self.TARGET._meta.verbose_name_plural)

        res = res.click('Add')
        form = res.form
        form['username'] = 'CHAR'
        res = form.submit().follow()

    def test_normal_update(self):
        self.TARGET.objects.get_or_create(pk=next(unique_id))
        res = self.app.get('/admin/', user='sax')
        res = res.click(self.TARGET._meta.verbose_name_plural)
        form = res.forms['changelist-form']
        form['form-0-username'] = 'CHAR'
        res = form.submit('_save').follow()
        self.assertTrue(self.TARGET.objects.filter(username='CHAR').exists())

    def test_concurrency_policy_abort(self):
        id = next(unique_id)
        self.TARGET.objects.get_or_create(pk=id)
        model_admin = site._registry[self.TARGET]
        with attributes((model_admin.__class__, 'list_editable_policy', CONCURRENCY_LIST_EDITABLE_POLICY_ABORT_ALL)):
            res = self.app.get('/admin/', user='sax')
            res = res.click(self.TARGET._meta.verbose_name_plural)
            self._create_conflict(id)

            form = res.forms['changelist-form']
            form['form-0-username'] = 'CHAR'

            with pytest.raises(RecordModifiedError):
                res = form.submit('_save')

            self.assertTrue(self.TARGET.objects.filter(username=SENTINEL).exists())
            self.assertFalse(self.TARGET.objects.filter(username='CHAR').exists())

    def test_concurrency_policy_silent(self):
        id = next(unique_id)
        self.TARGET.objects.get_or_create(pk=id)
        model_admin = site._registry[self.TARGET]
        with attributes((model_admin.__class__, 'list_editable_policy', CONCURRENCY_LIST_EDITABLE_POLICY_SILENT)):
            res = self.app.get('/admin/', user='sax')
            res = res.click(self.TARGET._meta.verbose_name_plural)
            self._create_conflict(id)

            form = res.forms['changelist-form']
            form['form-0-username'] = 'CHAR'
            res = form.submit('_save').follow()
            self.assertTrue(self.TARGET.objects.filter(username=SENTINEL).exists())
            self.assertFalse(self.TARGET.objects.filter(username='CHAR').exists())

    def test_message_user(self):
        id1 = next(unique_id)
        id2 = next(unique_id)
        self.TARGET.objects.get_or_create(pk=id1)
        self.TARGET.objects.get_or_create(pk=id2)
        res = self.app.get('/admin/', user='sax')
        res = res.click(self.TARGET._meta.verbose_name_plural)

        self._create_conflict(id1)

        form = res.forms['changelist-form']
        form['form-0-username'] = 'CHAR1'
        form['form-1-username'] = 'CHAR2'
        res = form.submit('_save').follow()

        messages = map(str, list(res.context['messages']))

        self.assertIn('Record with pk `%s` has been modified and was not updated' % id1,
                      messages)
        self.assertIn('1 %s was changed successfully.' % force_text(self.TARGET._meta.verbose_name),
                      messages)

    def test_message_user_no_changes(self):
        id = next(unique_id)
        self.TARGET.objects.get_or_create(pk=id)

        res = self.app.get('/admin/', user='sax')
        res = res.click(self.TARGET._meta.verbose_name_plural)

        self._create_conflict(id)

        form = res.forms['changelist-form']
        form['form-0-username'] = 'CHAR1'
        res = form.submit('_save').follow()

        messages = list(map(str, list(res.context['messages'])))

        self.assertIn('Record with pk `%s` has been modified and was not updated' % id, messages)
        self.assertEqual(len(messages), 1)

    def test_log_change(self):
        id = next(unique_id)
        self.TARGET.objects.get_or_create(pk=id)

        res = self.app.get('/admin/', user='sax')
        res = res.click(self.TARGET._meta.verbose_name_plural)
        log_filter = dict(user__username='sax',
                          content_type=ContentType.objects.get_for_model(self.TARGET))

        logs = list(LogEntry.objects.filter(**log_filter).values_list('pk', flat=True))

        self._create_conflict(id)

        form = res.forms['changelist-form']
        form['form-0-username'] = 'CHAR1'
        res = form.submit('_save').follow()
        new_logs = LogEntry.objects.filter(**log_filter).exclude(id__in=logs).exists()
        self.assertFalse(new_logs, "LogEntry created even if conflict error")
        transaction.rollback()

# class TestListEditableWithNoActions(TestListEditable):
#     TARGET = NoActionsConcurrentModel



from django.core.exceptions import ImproperlyConfigured, SuspiciousOperation
from django.forms.models import modelform_factory
from django.forms.widgets import HiddenInput, TextInput
from django.test import TestCase
from django.test.testcases import SimpleTestCase
from django.utils.encoding import smart_str
from django.utils.translation import ugettext as _

import pytest
from demo.models import Issue3TestModel, SimpleConcurrentModel

from concurrency.exceptions import VersionError
from concurrency.forms import (
    ConcurrentForm, VersionField, VersionFieldSigner, VersionWidget
)

__all__ = ['WidgetTest', 'FormFieldTest', 'ConcurrentFormTest']


class DummySigner():
    def sign(self, value):
        return smart_str(value)

    def unsign(self, signed_value):
        return smart_str(signed_value)


class WidgetTest(TestCase):
    def test(self):
        w = VersionWidget()
        self.assertHTMLEqual(w.render('ver', None),
                             '<input name="ver" type="hidden"/><div></div>')
        self.assertHTMLEqual(w.render('ver', 100),
                             '<input name="ver" type="hidden" value="100"/><div>100</div>')


class FormFieldTest(SimpleTestCase):
    def test_with_wrong_signer(self):
        with self.settings(CONCURRENCY_FIELD_SIGNER='invalid.Signer'):
            with pytest.raises(ImproperlyConfigured):
                VersionField()

    def test_with_dummy_signer(self):
        f = VersionField(signer=DummySigner())
        self.assertEqual(1, f.clean(1))
        self.assertEqual(1, f.clean('1'))
        self.assertEqual(0, f.clean(None))
        self.assertEqual(0, f.clean(''))
        self.assertRaises(VersionError, f.clean, 'aa:bb')
        self.assertRaises(VersionError, f.clean, 1.5)

    def test(self):
        f = VersionField()
        self.assertEqual(1, f.clean(VersionFieldSigner().sign(1)))
        self.assertEqual(1, f.clean(VersionFieldSigner().sign('1')))
        self.assertEqual(0, f.clean(None))
        self.assertEqual(0, f.clean(''))
        self.assertRaises(VersionError, f.clean, '100')
        self.assertRaises(VersionError, f.clean, VersionFieldSigner().sign(1.5))


class ConcurrentFormTest(TestCase):
    def test_version(self):
        Form = modelform_factory(SimpleConcurrentModel, ConcurrentForm, exclude=('char_field',))
        form = Form()
        self.assertIsInstance(form.fields['version'].widget, HiddenInput)

    def test_clean(self):
        pass

    def test_dummy_signer(self):
        obj, __ = Issue3TestModel.objects.get_or_create(username='aaa')
        Form = modelform_factory(Issue3TestModel,
                                 fields=('id', 'revision'),
                                 form=type('xxx', (ConcurrentForm,), {'revision': VersionField(signer=DummySigner())}))
        data = {'id': 1,
                'revision': obj.revision}
        form = Form(data, instance=obj)
        self.assertTrue(form.is_valid(), form.non_field_errors())

    def test_signer(self):
        Form = modelform_factory(Issue3TestModel, form=ConcurrentForm,
                                 exclude=('char_field',))
        form = Form({'username': 'aaa'})
        self.assertTrue(form.is_valid(), form.non_field_errors())

    def test_initial_value(self):
        Form = modelform_factory(SimpleConcurrentModel, type('xxx', (ConcurrentForm,), {}), exclude=('char_field',))
        form = Form({'username': 'aaa'})
        self.assertHTMLEqual(str(form['version']), '<input type="hidden" value="" name="version" id="id_version" />')
        self.assertTrue(form.is_valid(), form.non_field_errors())

    def test_initial_value_with_custom_signer(self):
        Form = modelform_factory(Issue3TestModel, exclude=('char_field',),
                                 form=type('xxx', (ConcurrentForm,),
                                           {'version': VersionField(signer=DummySigner())}))
        form = Form({'username': 'aaa'})
        self.assertHTMLEqual(str(form['version']), '<input type="hidden" value="" name="version" id="id_version" />')
        self.assertTrue(form.is_valid(), form.non_field_errors())

    def test_tamperig(self):
        obj, __ = Issue3TestModel.objects.get_or_create(username='aaa')
        Form = modelform_factory(Issue3TestModel, ConcurrentForm, exclude=('char_field',))
        data = {'username': 'aaa',
                'last_name': None,
                'date_field': None,
                'char_field': None,
                'version': 'abc',
                'id': 1,
                'revision': obj.revision}
        form = Form(data, instance=obj)
        self.assertRaises(SuspiciousOperation, form.is_valid)

    def test_custom_name(self):
        Form = modelform_factory(Issue3TestModel, ConcurrentForm, exclude=('char_field',))
        form = Form()
        self.assertIsInstance(form.fields['version'].widget, TextInput)
        self.assertIsInstance(form.fields['revision'].widget, HiddenInput)

    def test_save(self):
        obj, __ = Issue3TestModel.objects.get_or_create(username='aaa')

        obj_copy = Issue3TestModel.objects.get(pk=obj.pk)
        Form = modelform_factory(Issue3TestModel, ConcurrentForm,
                                 fields=('username', 'last_name', 'date_field',
                                         'char_field', 'version', 'id', 'revision'))
        data = {'username': 'aaa',
                'last_name': None,
                'date_field': None,
                'char_field': None,
                'version': 'abc',
                'id': 1,
                'revision': VersionFieldSigner().sign(obj.revision)}
        form = Form(data, instance=obj)
        obj_copy.save()  # save

        self.assertFalse(form.is_valid())
        self.assertIn(_('Record Modified'), form.non_field_errors())

    def test_is_valid(self):
        obj, __ = Issue3TestModel.objects.get_or_create(username='aaa')
        Form = modelform_factory(Issue3TestModel, ConcurrentForm,
                                 fields=('username', 'last_name', 'date_field',
                                         'char_field', 'version', 'id', 'revision'))
        data = {'username': 'aaa',
                'last_name': None,
                'date_field': None,
                'char_field': None,
                'version': 'abc',
                'id': 1,
                'revision': VersionFieldSigner().sign(obj.revision)}
        form = Form(data, instance=obj)
        obj.save()  # save again simulate concurrent editing
        self.assertRaises(ValueError, form.save)



# -*- coding: utf-8 -*-
import logging

import six
from django.core.management import call_command

import pytest
from mock import Mock

import concurrency.management.commands.triggers as command

logger = logging.getLogger(__name__)


@pytest.mark.django_db
def test_command_create(monkeypatch):
    out = six.StringIO()
    mock_create = Mock()
    mock_create.return_value = {'default': [['model', 'field', 'trigger']]}

    monkeypatch.setattr(command, 'create_triggers', mock_create)
    call_command('triggers', 'create', stdout=out)

    out.seek(0)
    output = out.read()
    assert output.find('Created trigger  for field') > 0
    assert mock_create.call_count == 1


@pytest.mark.django_db
def test_command_list():
    out = six.StringIO()
    call_command('triggers', 'list', stdout=out)
    out.seek(0)
    output = out.read()
    assert output.find('concurrency_demo_triggerconcurrentmodel_i')
    assert output.find('concurrency_demo_triggerconcurrentmodel_u')


@pytest.mark.django_db
def test_command_drop(monkeypatch):
    out = six.StringIO()
    mock_drop = Mock()
    mock_drop.return_value = {'default': [['model', 'field', 'trigger']]}

    monkeypatch.setattr(command, 'drop_triggers', mock_drop)
    call_command('triggers', 'drop', stdout=out)

    out.seek(0)
    output = out.read()
    assert output.find('Dropped   trigger') > 0
    assert mock_drop.call_count == 1



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import json
import logging
import os

from django.core.management import call_command
from six import StringIO

import pytest
from demo.models import SimpleConcurrentModel

from concurrency.api import disable_concurrency
from concurrency.exceptions import RecordModifiedError

logger = logging.getLogger(__name__)


@pytest.mark.django_db()
def test_dumpdata():
    SimpleConcurrentModel.objects.create()
    out = StringIO()
    call_command('dumpdata', 'demo', stdout=out)
    data = json.loads(out.getvalue())
    assert data


@pytest.mark.django_db(transaction=True)
def test_loaddata_fail():
    datafile = os.path.join(os.path.dirname(__file__), 'dumpdata.json')
    data = json.load(open(datafile, 'r'))
    pk = data[0]['pk']

    call_command('loaddata', datafile, stdout=StringIO())

    assert SimpleConcurrentModel.objects.get(id=pk).username == 'loaded'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('demo', '0002_conditionalversionmodelwithoutmeta'),
    ]

    operations = [
        migrations.CreateModel(
            name='Anything',
            fields=[
                ('id', models.AutoField(primary_key=True, verbose_name='ID', serialize=False, auto_created=True)),
                ('name', models.CharField(max_length=10)),
                ('a_relation', models.ForeignKey(to='demo.ConditionalVersionModelWithoutMeta')),
            ],
        ),
        migrations.AddField(
            model_name='conditionalversionmodelwithoutmeta',
            name='anythings',
            field=models.ManyToManyField(to='demo.Anything'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import logging

from django.contrib.auth.models import User

import pytest
from demo.models import ConditionalVersionModel, ConditionalVersionModelWithoutMeta

from concurrency.exceptions import RecordModifiedError
from concurrency.utils import refetch

logger = logging.getLogger(__name__)


@pytest.fixture
def user():
    return User.objects.get_or_create(username='username')[0]


@pytest.fixture
def instance(user):
    return ConditionalVersionModel.objects.get_or_create(field1='1',
                                                         user=user,
                                                         field2='1', field3='1')[0]


@pytest.fixture
def instance_no_meta(user):
    return ConditionalVersionModelWithoutMeta.objects.create(
        field1='1',
        user=user,
        field2='1', field3='1'
    )


@pytest.mark.django_db
def test_standard_save(instance):
    # only increment if checked field
    instance = refetch(instance)
    version1 = instance.get_concurrency_version()
    assert version1 == 1  # version2 > version1

    instance.field1 = '2'
    instance.save()
    version2 = instance.get_concurrency_version()
    assert version2 == 2  # version2 > version1

    instance.field3 = '3'
    instance.save()
    version3 = instance.get_concurrency_version()
    assert version3 == 2  # version3 == version2

    instance.user = None
    instance.save()
    version4 = instance.get_concurrency_version()
    assert version4 == 3  # version4 > version2


@pytest.mark.django_db(transaction=True)
def test_conflict(instance):
    # Scenario: batch change field present in ConcurrencyMeta.check_fields
    # the user is NOT ALLOWED to save
    batch_instance = instance.__class__.objects.get(pk=instance.pk)
    assert batch_instance.version == instance.version

    batch_instance.field1 = 'aaaa'
    batch_instance.save()

    with pytest.raises(RecordModifiedError):
        instance.save()


@pytest.mark.django_db(transaction=True)
def test_save_allowed(instance):
    # Scenario: batch change field NOT present in ConcurrencyMeta.check_fields
    # the user is ALLOWED to save
    batch_instance = instance.__class__.objects.get(pk=instance.pk)
    assert batch_instance.version == instance.version
    instance = refetch(instance)
    batch_instance = refetch(instance)

    batch_instance.field3 = 'aaaa'
    batch_instance.save()
    instance.save()


@pytest.mark.django_db(transaction=True)
def test_conflict_no_meta(instance_no_meta):
    # Scenario: batch change any field,
    # the user is NOT ALLOWED to save
    batch_instance = instance_no_meta.__class__.objects.get(pk=instance_no_meta.pk)
    assert batch_instance.version == instance_no_meta.version

    batch_instance.field1 = 'aaaa'
    batch_instance.save()

    with pytest.raises(RecordModifiedError):
        instance_no_meta.save()



#!/usr/bin/env python
import imp
import os
import sys

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))
init = os.path.join(ROOT, 'src', 'concurrency', '__init__.py')
app = imp.load_source('concurrency', init)

reqs = 'install.py%d.pip' % sys.version_info[0]

rel = lambda fname: os.path.join(os.path.dirname(__file__),
                                 'src',
                                 'requirements', fname)


def fread(fname):
    return open(rel(fname)).read()

install_requires = fread('install.pip')
test_requires = fread('testing.pip')
dev_requires = fread('develop.pip')

base_url = 'https://github.com/saxix/django-concurrency/'


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['tests']
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        import sys

        sys.path.insert(0, os.path.join(ROOT, 'tests', 'demoapp'))
        errno = pytest.main(self.test_args)
        sys.exit(errno)

setup(
    name=app.NAME,
    version=app.get_version(),
    url='https://github.com/saxix/django-concurrency',
    author='Stefano Apostolico',
    author_email='s.apostolico@gmail.com',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    include_package_data=True,
    description='Optimistic lock implementation for Django. Prevents users from doing concurrent editing.',
    long_description=open('README.rst').read(),
    license='MIT License',
    keywords='django',
    setup_requires=['pytest-runner', ],
    install_requires=install_requires,
    tests_require='django\n' + test_requires,
    extras_require={'test': test_requires,
                    'dev': test_requires + dev_requires},
    # cmdclass={'test': PyTest},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    platforms=['any'],
)



# -*- coding: utf-8 -*-
import logging

from django.db import connections

import pytest
from demo.models import DropTriggerConcurrentModel, TriggerConcurrentModel  # noqa

from concurrency.triggers import drop_triggers, factory

logger = logging.getLogger(__name__)


@pytest.mark.django_db
def test_list_triggers():
    conn = connections['default']

    assert factory(conn).get_list() == [
        u'concurrency_demo_droptriggerconcurrentmodel_version',
        u'concurrency_demo_triggerconcurrentmodel_version']


@pytest.mark.django_db
def test_get_trigger(monkeypatch):
    conn = connections['default']
    f = factory(conn)
    version_field = TriggerConcurrentModel._concurrencymeta.field
    trigger = f.get_trigger(version_field)
    assert trigger == 'concurrency_demo_triggerconcurrentmodel_version'

    monkeypatch.setattr(version_field, '_trigger_name', 'aaa')
    assert f.get_trigger(version_field) is None


@pytest.mark.skipif('connections["default"].vendor=="mysql"',
                    reason="Mysql is not able to drop tringger inside trasaction")
@pytest.mark.django_db
def test_drop_trigger():
    conn = connections['default']
    f = [f for f in DropTriggerConcurrentModel._meta.fields if f.name == 'version'][0]
    ret = factory(conn).drop(f)
    assert ret == [u'concurrency_demo_droptriggerconcurrentmodel_version']
    assert factory(conn).get_list() == [u'concurrency_demo_triggerconcurrentmodel_version']


@pytest.mark.skipif('connections["default"].vendor=="mysql"',
                    reason="Mysql is not able to drop tringger inside trasaction")
@pytest.mark.django_db
def test_drop_triggers(db):
    conn = connections['default']
    ret = drop_triggers('default')
    assert sorted([i[0].__name__ for i in ret['default']]) == ['DropTriggerConcurrentModel',
                                                               'TriggerConcurrentModel']
    assert factory(conn).get_list() == []



# -*- coding: utf-8 -*-
import re

import django
from django.contrib.admin.sites import site
from django.contrib.auth.models import User
from django.core.management import call_command
from django.http import QueryDict
from django.test.client import RequestFactory
from django.test.testcases import SimpleTestCase
from django.utils.encoding import force_text

import pytest
from demo.admin import ActionsModelAdmin, admin_register
from demo.base import AdminTestCase
from demo.models import ListEditableConcurrentModel, ReversionConcurrentModel
from demo.util import attributes, unique_id

from concurrency.admin import ConcurrentModelAdmin
from concurrency.config import CONCURRENCY_LIST_EDITABLE_POLICY_SILENT
from concurrency.forms import ConcurrentForm
from concurrency.templatetags.concurrency import identity
from concurrency.utils import refetch


def get_fake_request(params):
    u, __ = User.objects.get_or_create(username='sax')
    setattr(u, 'is_authenticated()', True)
    setattr(u, 'selected_office', False)

    request = RequestFactory().request()
    request.user = u

    querydict = QueryDict(params)
    request.POST = querydict

    return request


class TestIssue16(AdminTestCase):
    def test_concurrency(self):
        id = 1
        admin_register(ListEditableConcurrentModel, ActionsModelAdmin)
        model_admin = site._registry[ListEditableConcurrentModel]
        with attributes((ConcurrentModelAdmin, 'list_editable_policy', CONCURRENCY_LIST_EDITABLE_POLICY_SILENT),
                        (ConcurrentModelAdmin, 'form', ConcurrentForm), ):
            obj, __ = ListEditableConcurrentModel.objects.get_or_create(pk=id)
            request1 = get_fake_request('pk=%s&_concurrency_version_1=2' % id)

            model_admin.save_model(request1, obj, None, True)

            self.assertIn(obj.pk, model_admin._get_conflicts(request1))

            obj = refetch(obj)
            request2 = get_fake_request('pk=%s&_concurrency_version_1=%s' % (id, obj.version))
            model_admin.save_model(request2, obj, None, True)
            self.assertNotIn(obj.pk, model_admin._get_conflicts(request2))


class TestIssue18(SimpleTestCase):
    def test_identity_tag(self):
        id = next(unique_id)

        obj = ListEditableConcurrentModel(pk=id)
        self.assertTrue(re.match(r"^%s,\d+$" % id, identity(obj)))

        g = User(username='UserTest', pk=3)
        self.assertEqual(identity(g), force_text(g.pk))


@pytest.mark.skipif(django.VERSION[:2] >= (1, 9), reason="Skip django>=1.9")
@pytest.mark.django_db()
def test_issue_53(admin_client):
    pytest.importorskip("reversion")
    import reversion as revisions

    with revisions.create_revision():
        instance = ReversionConcurrentModel.objects.create()
    pk = instance.pk

    with revisions.create_revision():
        instance.delete()

    version_list = revisions.get_deleted(ReversionConcurrentModel)
    deleted_pk = version_list[0].pk
    admin_client.post('/admin/demo/reversionconcurrentmodel/recover/{}/'.format(deleted_pk),
                      {'username': 'aaaa'})
    assert ReversionConcurrentModel.objects.filter(id=pk).exists()



# -*- coding: utf-8 -*-
from django.core.signals import request_started
from django.db import IntegrityError, connections

import mock
import pytest
from demo.models import TriggerConcurrentModel
# Register an event to reset saved queries when a Django request is started.
from demo.util import nextname

from concurrency.exceptions import RecordModifiedError
from concurrency.utils import refetch


def reset_queries(**kwargs):
    for conn in connections.all():
        conn.queries = []


class CaptureQueriesContext(object):
    """
    Context manager that captures queries executed by the specified connection.
    """

    def __init__(self, connection):
        self.connection = connection

    def __iter__(self):
        return iter(self.captured_queries)

    def __getitem__(self, index):
        return self.captured_queries[index]

    def __len__(self):
        return len(self.captured_queries)

    @property
    def captured_queries(self):
        return self.connection.queries[self.initial_queries:self.final_queries]

    def __enter__(self):
        self.use_debug_cursor = self.connection.use_debug_cursor
        self.connection.use_debug_cursor = True
        self.initial_queries = len(self.connection.queries)
        self.final_queries = None
        request_started.disconnect(reset_queries)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.connection.use_debug_cursor = self.use_debug_cursor
        request_started.connect(reset_queries)
        if exc_type is not None:
            return
        self.final_queries = len(self.connection.queries)


@pytest.mark.django_db
def test_trigger():
    instance = TriggerConcurrentModel()
    assert instance.pk is None
    assert instance.version == 0

    instance.username = next(nextname)
    instance.save()  # insert
    instance = refetch(instance)
    assert instance.version == 1

    instance.username = next(nextname)
    instance.save()  # update
    assert instance.version == 2

    instance.username = next(nextname)
    instance.save()  # update
    assert instance.version == 3

    instance.username = next(nextname)
    instance.save(refetch=True)  # update
    assert instance.version == 4

    copy = refetch(instance)
    copy.save()

    with pytest.raises(RecordModifiedError):
        instance.save()


@pytest.mark.django_db
def test_trigger_do_not_increase_version_if_error():
    instance = TriggerConcurrentModel()
    assert instance.pk is None
    assert instance.version == 0
    with mock.patch('demo.models.TriggerConcurrentModel.save', side_effect=IntegrityError):
        with pytest.raises(IntegrityError):
            instance.save()

    assert instance.version == 0
